#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from openstackclient.tests.unit.volume.v2 import fakes as volume_fakes
from openstackclient.volume.v2 import volume_backend


class TestShowVolumeCapability(volume_fakes.TestVolume):
    """Test backend capability functionality."""

    # The capability to be listed
    capability = volume_fakes.create_one_capability()

    def setUp(self):
        super().setUp()

        # Get a shortcut to the capability Mock
        self.capability_mock = self.app.client_manager.volume.capabilities
        self.capability_mock.get.return_value = self.capability

        # Get the command object to test
        self.cmd = volume_backend.ShowCapability(self.app, None)

    def test_capability_show(self):
        arglist = [
            'fake',
        ]
        verifylist = [
            ('host', 'fake'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'Title',
            'Key',
            'Type',
            'Description',
        ]

        # confirming if all expected columns are present in the result.
        self.assertEqual(expected_columns, columns)

        capabilities = [
            'Compression',
            'Replication',
            'QoS',
            'Thin Provisioning',
        ]

        # confirming if all expected values are present in the result.
        for cap in data:
            self.assertIn(cap[0], capabilities)

        # checking if proper call was made to get capabilities
        self.capability_mock.get.assert_called_with(
            'fake',
        )


class TestListVolumePool(volume_fakes.TestVolume):
    """Tests for volume backend pool listing."""

    # The pool to be listed
    pools = volume_fakes.create_one_pool()

    def setUp(self):
        super().setUp()

        self.pool_mock = self.app.client_manager.volume.pools
        self.pool_mock.list.return_value = [self.pools]

        # Get the command object to test
        self.cmd = volume_backend.ListPool(self.app, None)

    def test_pool_list(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'Name',
        ]

        # confirming if all expected columns are present in the result.
        self.assertEqual(expected_columns, columns)

        datalist = ((
            self.pools.name,
        ), )

        # confirming if all expected values are present in the result.
        self.assertEqual(datalist, tuple(data))

        # checking if proper call was made to list pools
        self.pool_mock.list.assert_called_with(
            detailed=False,
        )

        # checking if long columns are present in output
        self.assertNotIn("total_volumes", columns)
        self.assertNotIn("storage_protocol", columns)

    def test_service_list_with_long_option(self):
        arglist = [
            '--long'
        ]
        verifylist = [
            ('long', True)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'Name',
            'Protocol',
            'Thick',
            'Thin',
            'Volumes',
            'Capacity',
            'Allocated',
            'Max Over Ratio',
        ]

        # confirming if all expected columns are present in the result.
        self.assertEqual(expected_columns, columns)

        datalist = ((
            self.pools.name,
            self.pools.storage_protocol,
            self.pools.thick_provisioning_support,
            self.pools.thin_provisioning_support,
            self.pools.total_volumes,
            self.pools.total_capacity_gb,
            self.pools.allocated_capacity_gb,
            self.pools.max_over_subscription_ratio,
        ), )

        # confirming if all expected values are present in the result.
        self.assertEqual(datalist, tuple(data))

        self.pool_mock.list.assert_called_with(
            detailed=True,
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from openstackclient.common import configuration
from openstackclient.tests.functional import base


BASIC_CONFIG_HEADERS = ['Field', 'Value']


class ConfigurationTests(base.TestCase):
    """Functional test for configuration."""

    def test_configuration_show(self):

        # Test show without option
        raw_output = self.openstack('configuration show')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, BASIC_CONFIG_HEADERS)

        cmd_output = self.openstack('configuration show', parse_output=True)
        self.assertEqual(
            configuration.REDACTED,
            cmd_output['auth.password']
        )
        self.assertIn(
            'auth.password',
            cmd_output.keys(),
        )

        # Test show --mask
        cmd_output = self.openstack(
            'configuration show --mask', parse_output=True,
        )
        self.assertEqual(
            configuration.REDACTED,
            cmd_output['auth.password']
        )

        # Test show --unmask
        cmd_output = self.openstack(
            'configuration show --unmask', parse_output=True,
        )
        # If we are using os-client-config, this will not be set.  Rather than
        # parse clouds.yaml to get the right value, just make sure
        # we are not getting redacted.
        passwd = os.environ.get('OS_PASSWORD')
        if passwd:
            self.assertEqual(
                passwd,
                cmd_output['auth.password']
            )
        else:
            self.assertNotEqual(
                configuration.REDACTED,
                cmd_output['auth.password']
            )


class ConfigurationTestsNoAuth(base.TestCase):
    """Functional test for configuration with no auth"""

    def test_configuration_show(self):

        # Test show without option
        raw_output = self.openstack(
            'configuration show',
            cloud=None,
        )
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, BASIC_CONFIG_HEADERS)

        cmd_output = self.openstack(
            'configuration show',
            cloud=None,
            parse_output=True,
        )
        self.assertNotIn(
            'auth.password',
            cmd_output,
        )



#   Copyright 2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Flavor action implementations"""

import logging

from openstack import exceptions as sdk_exceptions
from openstack import utils as sdk_utils
from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


LOG = logging.getLogger(__name__)


_formatters = {
    'extra_specs': format_columns.DictColumn,
    'properties': format_columns.DictColumn
}


def _get_flavor_columns(item):
    # To maintain backwards compatibility we need to rename sdk props to
    # whatever OSC was using before
    column_map = {
        'extra_specs': 'properties',
        'ephemeral': 'OS-FLV-EXT-DATA:ephemeral',
        'is_disabled': 'OS-FLV-DISABLED:disabled',
        'is_public': 'os-flavor-access:is_public'

    }
    hidden_columns = ['links', 'location', 'original_name']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns)


class CreateFlavor(command.ShowOne):
    _description = _("Create new flavor")

    def get_parser(self, prog_name):
        parser = super(CreateFlavor, self).get_parser(prog_name)
        parser.add_argument(
            "name",
            metavar="<flavor-name>",
            help=_("New flavor name")
        )
        parser.add_argument(
            "--id",
            metavar="<id>",
            help=_("Unique flavor ID")
        )
        parser.add_argument(
            "--ram",
            type=int,
            metavar="<size-mb>",
            default=256,
            help=_("Memory size in MB (default 256M)")
        )
        parser.add_argument(
            "--disk",
            type=int,
            metavar="<size-gb>",
            default=0,
            help=_("Disk size in GB (default 0G)")
        )
        parser.add_argument(
            "--ephemeral",
            type=int,
            metavar="<size-gb>",
            default=0,
            help=_("Ephemeral disk size in GB (default 0G)")
        )
        parser.add_argument(
            "--swap",
            type=int,
            metavar="<size-mb>",
            default=0,
            help=_("Additional swap space size in MB (default 0M)")
        )
        parser.add_argument(
            "--vcpus",
            type=int,
            metavar="<vcpus>",
            default=1,
            help=_("Number of vcpus (default 1)")
        )
        parser.add_argument(
            "--rxtx-factor",
            type=float,
            metavar="<factor>",
            default=1.0,
            help=_("RX/TX factor (default 1.0)")
        )
        public_group = parser.add_mutually_exclusive_group()
        public_group.add_argument(
            "--public",
            dest="public",
            action="store_true",
            default=True,
            help=_("Flavor is available to other projects (default)")
        )
        public_group.add_argument(
            "--private",
            dest="public",
            action="store_false",
            help=_("Flavor is not available to other projects")
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            action=parseractions.KeyValueAction,
            dest="properties",
            help=_("Property to add for this flavor "
                   "(repeat option to set multiple properties)")
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_("Allow <project> to access private flavor (name or ID) "
                   "(Must be used with --private option)"),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_("Description for the flavor.(Supported by API versions "
                   "'2.55' - '2.latest'")
        )
        identity_common.add_project_domain_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.sdk_connection.compute
        identity_client = self.app.client_manager.identity

        if parsed_args.project and parsed_args.public:
            msg = _("--project is only allowed with --private")
            raise exceptions.CommandError(msg)

        args = {
            'name': parsed_args.name,
            'ram': parsed_args.ram,
            'vcpus': parsed_args.vcpus,
            'disk': parsed_args.disk,
            'id': parsed_args.id,
            'ephemeral': parsed_args.ephemeral,
            'swap': parsed_args.swap,
            'rxtx_factor': parsed_args.rxtx_factor,
            'is_public': parsed_args.public,
        }

        if parsed_args.description:
            if not sdk_utils.supports_microversion(compute_client, '2.55'):
                msg = _(
                    'The --description parameter requires server support for '
                    'API microversion 2.55'
                )
                raise exceptions.CommandError(msg)

            args['description'] = parsed_args.description

        flavor = compute_client.create_flavor(**args)

        if parsed_args.project:
            try:
                project_id = identity_common.find_project(
                    identity_client,
                    parsed_args.project,
                    parsed_args.project_domain,
                ).id
                compute_client.flavor_add_tenant_access(
                    flavor.id, project_id)
            except Exception as e:
                msg = _("Failed to add project %(project)s access to "
                        "flavor: %(e)s")
                LOG.error(msg, {'project': parsed_args.project, 'e': e})
        if parsed_args.properties:
            try:
                flavor = compute_client.create_flavor_extra_specs(
                    flavor, parsed_args.properties)
            except Exception as e:
                LOG.error(_("Failed to set flavor properties: %s"), e)

        display_columns, columns = _get_flavor_columns(flavor)
        data = utils.get_dict_properties(flavor, columns,
                                         formatters=_formatters)

        return (display_columns, data)


class DeleteFlavor(command.Command):
    _description = _("Delete flavor(s)")

    def get_parser(self, prog_name):
        parser = super(DeleteFlavor, self).get_parser(prog_name)
        parser.add_argument(
            "flavor",
            metavar="<flavor>",
            nargs='+',
            help=_("Flavor(s) to delete (name or ID)")
        )
        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.sdk_connection.compute
        result = 0
        for f in parsed_args.flavor:
            try:
                flavor = compute_client.find_flavor(f, ignore_missing=False)
                compute_client.delete_flavor(flavor.id)
            except Exception as e:
                result += 1
                LOG.error(_("Failed to delete flavor with name or "
                          "ID '%(flavor)s': %(e)s"), {'flavor': f, 'e': e})

        if result > 0:
            total = len(parsed_args.flavor)
            msg = (_("%(result)s of %(total)s flavors failed "
                   "to delete.") % {'result': result, 'total': total})
            raise exceptions.CommandError(msg)


class ListFlavor(command.Lister):
    _description = _("List flavors")

    def get_parser(self, prog_name):
        parser = super(ListFlavor, self).get_parser(prog_name)
        public_group = parser.add_mutually_exclusive_group()
        public_group.add_argument(
            "--public",
            dest="public",
            action="store_true",
            default=True,
            help=_("List only public flavors (default)")
        )
        public_group.add_argument(
            "--private",
            dest="public",
            action="store_false",
            help=_("List only private flavors")
        )
        public_group.add_argument(
            "--all",
            dest="all",
            action="store_true",
            default=False,
            help=_("List all flavors, whether public or private")
        )
        parser.add_argument(
            '--min-disk',
            type=int,
            metavar='<min-disk>',
            help=_('Filters the flavors by a minimum disk space, in GiB.'),
        )
        parser.add_argument(
            '--min-ram',
            type=int,
            metavar='<min-ram>',
            help=_('Filters the flavors by a minimum RAM, in MiB.'),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_("List additional fields in output")
        )
        parser.add_argument(
            '--marker',
            metavar="<flavor-id>",
            help=_("The last flavor ID of the previous page")
        )
        parser.add_argument(
            '--limit',
            type=int,
            metavar='<num-flavors>',
            help=_(
                'Maximum number of flavors to display. This is also '
                'configurable on the server. The actual limit used will be '
                'the lower of the user-supplied value and the server '
                'configuration-derived value'
            ),
        )
        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.sdk_connection.compute
        # is_public is ternary - None means give all flavors,
        # True is public only and False is private only
        # By default Nova assumes True and gives admins public flavors
        # and flavors from their own projects only.
        is_public = None if parsed_args.all else parsed_args.public

        query_attrs = {
            'is_public': is_public
        }

        if parsed_args.marker:
            query_attrs['marker'] = parsed_args.marker

        if parsed_args.limit:
            query_attrs['limit'] = parsed_args.limit

        if parsed_args.limit or parsed_args.marker:
            # User passed explicit pagination request, switch off SDK
            # pagination
            query_attrs['paginated'] = False

        if parsed_args.min_disk:
            query_attrs['min_disk'] = parsed_args.min_disk

        if parsed_args.min_ram:
            query_attrs['min_ram'] = parsed_args.min_ram

        data = list(compute_client.flavors(**query_attrs))
        # Even if server supports 2.61 some policy might stop it sending us
        # extra_specs. So try to fetch them if they are absent
        for f in data:
            if parsed_args.long and not f.extra_specs:
                compute_client.fetch_flavor_extra_specs(f)

        columns = (
            "id",
            "name",
            "ram",
            "disk",
            "ephemeral",
            "vcpus",
            "is_public"
        )
        if parsed_args.long:
            columns += (
                "swap",
                "rxtx_factor",
                "extra_specs",
            )

        column_headers = (
            "ID",
            "Name",
            "RAM",
            "Disk",
            "Ephemeral",
            "VCPUs",
            "Is Public"
        )
        if parsed_args.long:
            column_headers += (
                "Swap",
                "RXTX Factor",
                "Properties",
            )

        return (
            column_headers,
            (
                utils.get_item_properties(s, columns, formatters=_formatters)
                for s in data
            ),
        )


class SetFlavor(command.Command):
    _description = _("Set flavor properties")

    def get_parser(self, prog_name):
        parser = super(SetFlavor, self).get_parser(prog_name)
        parser.add_argument(
            "flavor",
            metavar="<flavor>",
            help=_("Flavor to modify (name or ID)")
        )
        parser.add_argument(
            "--no-property",
            action="store_true",
            help=_("Remove all properties from this flavor "
                   "(specify both --no-property and --property"
                   " to remove the current properties before setting"
                   " new properties.)"),
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            action=parseractions.KeyValueAction,
            dest="properties",
            help=_("Property to add or modify for this flavor "
                   "(repeat option to set multiple properties)")
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_('Set flavor access to project (name or ID) '
                   '(admin only)'),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_("Set description for the flavor.(Supported by API "
                   "versions '2.55' - '2.latest'")
        )

        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.sdk_connection.compute
        identity_client = self.app.client_manager.identity

        try:
            flavor = compute_client.find_flavor(
                parsed_args.flavor,
                get_extra_specs=True,
                ignore_missing=False)
        except sdk_exceptions.ResourceNotFound as e:
            raise exceptions.CommandError(e.message)

        if parsed_args.description:
            if not sdk_utils.supports_microversion(compute_client, '2.55'):
                msg = _(
                    'The --description parameter requires server support for '
                    'API microversion 2.55'
                )
                raise exceptions.CommandError(msg)

            compute_client.update_flavor(
                flavor=flavor.id, description=parsed_args.description)

        result = 0
        if parsed_args.no_property:
            try:
                for key in flavor.extra_specs.keys():
                    compute_client.delete_flavor_extra_specs_property(
                        flavor.id, key)
            except Exception as e:
                LOG.error(_("Failed to clear flavor properties: %s"), e)
                result += 1

        if parsed_args.properties:
            try:
                compute_client.create_flavor_extra_specs(
                    flavor.id, parsed_args.properties)
            except Exception as e:
                LOG.error(_("Failed to set flavor properties: %s"), e)
                result += 1

        if parsed_args.project:
            try:
                if flavor.is_public:
                    msg = _("Cannot set access for a public flavor")
                    raise exceptions.CommandError(msg)
                else:
                    project_id = identity_common.find_project(
                        identity_client,
                        parsed_args.project,
                        parsed_args.project_domain,
                    ).id
                    compute_client.flavor_add_tenant_access(
                        flavor.id, project_id)
            except Exception as e:
                LOG.error(_("Failed to set flavor access to project: %s"), e)
                result += 1

        if result > 0:
            raise exceptions.CommandError(_("Command Failed: One or more of"
                                            " the operations failed"))


class ShowFlavor(command.ShowOne):
    _description = _("Display flavor details")

    def get_parser(self, prog_name):
        parser = super(ShowFlavor, self).get_parser(prog_name)
        parser.add_argument(
            "flavor",
            metavar="<flavor>",
            help=_("Flavor to display (name or ID)")
        )
        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.sdk_connection.compute
        flavor = compute_client.find_flavor(
            parsed_args.flavor, get_extra_specs=True, ignore_missing=False)

        access_projects = None
        # get access projects list of this flavor
        if not flavor.is_public:
            try:
                flavor_access = compute_client.get_flavor_access(
                    flavor=flavor.id)
                access_projects = [
                    utils.get_field(access, 'tenant_id')
                    for access in flavor_access]
            except Exception as e:
                msg = _("Failed to get access projects list "
                        "for flavor '%(flavor)s': %(e)s")
                LOG.error(msg, {'flavor': parsed_args.flavor, 'e': e})

        # Since we need to inject "access_project_id" into resource - convert
        # it to dict and treat it respectively
        flavor = flavor.to_dict()
        flavor['access_project_ids'] = access_projects

        display_columns, columns = _get_flavor_columns(flavor)
        data = utils.get_dict_properties(
            flavor, columns, formatters=_formatters)

        return (display_columns, data)


class UnsetFlavor(command.Command):
    _description = _("Unset flavor properties")

    def get_parser(self, prog_name):
        parser = super(UnsetFlavor, self).get_parser(prog_name)
        parser.add_argument(
            "flavor",
            metavar="<flavor>",
            help=_("Flavor to modify (name or ID)")
        )
        parser.add_argument(
            "--property",
            metavar="<key>",
            action='append',
            dest="properties",
            help=_("Property to remove from flavor "
                   "(repeat option to unset multiple properties)")
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_('Remove flavor access from project (name or ID) '
                   '(admin only)'),
        )
        identity_common.add_project_domain_option_to_parser(parser)

        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.sdk_connection.compute
        identity_client = self.app.client_manager.identity

        try:
            flavor = compute_client.find_flavor(
                parsed_args.flavor,
                get_extra_specs=True,
                ignore_missing=False)
        except sdk_exceptions.ResourceNotFound as e:
            raise exceptions.CommandError(_(e.message))

        result = 0
        if parsed_args.properties:
            for key in parsed_args.properties:
                try:
                    compute_client.delete_flavor_extra_specs_property(
                        flavor.id, key)
                except sdk_exceptions.SDKException as e:
                    LOG.error(_("Failed to unset flavor property: %s"), e)
                    result += 1

        if parsed_args.project:
            try:
                if flavor.is_public:
                    msg = _("Cannot remove access for a public flavor")
                    raise exceptions.CommandError(msg)

                project_id = identity_common.find_project(
                    identity_client,
                    parsed_args.project,
                    parsed_args.project_domain,
                ).id
                compute_client.flavor_remove_tenant_access(
                    flavor.id, project_id)
            except Exception as e:
                LOG.error(_("Failed to remove flavor access from project: %s"),
                          e)
                result += 1

        if result > 0:
            raise exceptions.CommandError(_("Command Failed: One or more of"
                                            " the operations failed"))



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import logging

from osc_lib import exceptions
from osc_lib import utils

from openstackclient.i18n import _


LOG = logging.getLogger(__name__)

DEFAULT_API_VERSION = '3'
API_VERSION_OPTION = 'os_volume_api_version'
API_NAME = "volume"
API_VERSIONS = {
    "1": "cinderclient.v1.client.Client",
    "2": "cinderclient.v2.client.Client",
    "3": "cinderclient.v3.client.Client",
}

# Save the microversion if in use
_volume_api_version = None


def make_client(instance):
    """Returns a volume service client."""

    # Defer client imports until we actually need them
    from cinderclient import extension
    from cinderclient.v3.contrib import list_extensions
    from cinderclient.v3 import volume_snapshots
    from cinderclient.v3 import volumes

    # Check whether the available cinderclient supports v1 or v2
    try:
        from cinderclient.v1 import services  # noqa
    except Exception:
        del API_VERSIONS['1']
    try:
        from cinderclient.v2 import services  # noqa
    except Exception:
        del API_VERSIONS['2']

    if _volume_api_version is not None:
        version = _volume_api_version
    else:
        version = instance._api_version[API_NAME]
        from cinderclient import api_versions
        # convert to APIVersion object
        version = api_versions.get_api_version(version)

    if version.ver_major == '1':
        # Monkey patch for v1 cinderclient
        volumes.Volume.NAME_ATTR = 'display_name'
        volume_snapshots.Snapshot.NAME_ATTR = 'display_name'

    volume_client = utils.get_client_class(
        API_NAME,
        version.ver_major,
        API_VERSIONS
    )
    LOG.debug('Instantiating volume client: %s', volume_client)

    # Set client http_log_debug to True if verbosity level is high enough
    http_log_debug = utils.get_effective_log_level() <= logging.DEBUG

    extensions = [extension.Extension('list_extensions', list_extensions)]

    # Remember interface only if it is set
    kwargs = utils.build_kwargs_dict('endpoint_type', instance.interface)

    endpoint_override = instance.sdk_connection.config.get_endpoint(
        'block-storage')

    client = volume_client(
        session=instance.session,
        extensions=extensions,
        http_log_debug=http_log_debug,
        region_name=instance.region_name,
        endpoint_override=endpoint_override,
        api_version=version,
        **kwargs
    )

    return client


def build_option_parser(parser):
    """Hook to add global options"""
    parser.add_argument(
        '--os-volume-api-version',
        metavar='<volume-api-version>',
        default=utils.env('OS_VOLUME_API_VERSION'),
        help=_('Volume API version, default=%s '
               '(Env: OS_VOLUME_API_VERSION)') % DEFAULT_API_VERSION
    )
    return parser


def check_api_version(check_version):
    """Validate version supplied by user

    Returns:

    * True if version is OK
    * False if the version has not been checked and the previous plugin
      check should be performed
    * throws an exception if the version is no good
    """

    # Defer client imports until we actually need them
    from cinderclient import api_versions

    global _volume_api_version

    # Copy some logic from novaclient 3.3.0 for basic version detection
    # NOTE(dtroyer): This is only enough to resume operations using API
    # version 3.0 or any valid version supplied by the user.
    _volume_api_version = api_versions.get_api_version(check_version)

    # Bypass X.latest format microversion
    if not _volume_api_version.is_latest():
        if _volume_api_version > api_versions.APIVersion("3.0"):
            if not _volume_api_version.matches(
                api_versions.MIN_VERSION,
                api_versions.MAX_VERSION,
            ):
                msg = _("versions supported by client: %(min)s - %(max)s") % {
                    "min": api_versions.MIN_VERSION,
                    "max": api_versions.MAX_VERSION,
                }
                raise exceptions.CommandError(msg)

            return True

    return False



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import copy
from unittest import mock

from cinderclient import api_versions
from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.tests.unit.volume.v2 import fakes as volume_fakes
from openstackclient.volume.v3 import volume


class TestVolumeSummary(volume_fakes.TestVolume):

    columns = [
        'Total Count',
        'Total Size',
    ]

    def setUp(self):
        super().setUp()

        self.volumes_mock = self.app.client_manager.volume.volumes
        self.volumes_mock.reset_mock()
        self.mock_vol_1 = volume_fakes.create_one_volume()
        self.mock_vol_2 = volume_fakes.create_one_volume()
        self.return_dict = {
            'volume-summary': {
                'total_count': 2,
                'total_size': self.mock_vol_1.size + self.mock_vol_2.size}}
        self.volumes_mock.summary.return_value = self.return_dict

        # Get the command object to test
        self.cmd = volume.VolumeSummary(self.app, None)

    def test_volume_summary(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.12')
        arglist = [
            '--all-projects',
        ]
        verifylist = [
            ('all_projects', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volumes_mock.summary.assert_called_once_with(
            all_tenants=True,
        )

        self.assertEqual(self.columns, columns)

        datalist = (
            2,
            self.mock_vol_1.size + self.mock_vol_2.size)
        self.assertCountEqual(datalist, tuple(data))

    def test_volume_summary_pre_312(self):
        arglist = [
            '--all-projects',
        ]
        verifylist = [
            ('all_projects', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args)
        self.assertIn(
            '--os-volume-api-version 3.12 or greater is required',
            str(exc))

    def test_volume_summary_with_metadata(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.36')

        combine_meta = {**self.mock_vol_1.metadata, **self.mock_vol_2.metadata}
        meta_dict = copy.deepcopy(self.return_dict)
        meta_dict['volume-summary']['metadata'] = combine_meta
        self.volumes_mock.summary.return_value = meta_dict

        new_cols = copy.deepcopy(self.columns)
        new_cols.extend(['Metadata'])

        arglist = [
            '--all-projects',
        ]
        verifylist = [
            ('all_projects', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volumes_mock.summary.assert_called_once_with(
            all_tenants=True,
        )

        self.assertEqual(new_cols, columns)

        datalist = (
            2,
            self.mock_vol_1.size + self.mock_vol_2.size,
            format_columns.DictColumn(combine_meta))
        self.assertCountEqual(datalist, tuple(data))


class TestVolumeRevertToSnapshot(volume_fakes.TestVolume):

    def setUp(self):
        super().setUp()

        self.volumes_mock = self.app.client_manager.volume.volumes
        self.volumes_mock.reset_mock()
        self.snapshots_mock = self.app.client_manager.volume.volume_snapshots
        self.snapshots_mock.reset_mock()
        self.mock_volume = volume_fakes.create_one_volume()
        self.mock_snapshot = volume_fakes.create_one_snapshot(
            attrs={'volume_id': self.volumes_mock.id})

        # Get the command object to test
        self.cmd = volume.VolumeRevertToSnapshot(self.app, None)

    def test_volume_revert_to_snapshot_pre_340(self):
        arglist = [
            self.mock_snapshot.id,
        ]
        verifylist = [
            ('snapshot', self.mock_snapshot.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args)
        self.assertIn(
            '--os-volume-api-version 3.40 or greater is required',
            str(exc))

    def test_volume_revert_to_snapshot(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.40')
        arglist = [
            self.mock_snapshot.id,
        ]
        verifylist = [
            ('snapshot', self.mock_snapshot.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = [self.mock_snapshot, self.mock_volume]
        with mock.patch.object(utils, 'find_resource',
                               side_effect=find_mock_result) as find_mock:
            self.cmd.take_action(parsed_args)

            self.volumes_mock.revert_to_snapshot.assert_called_once_with(
                volume=self.mock_volume,
                snapshot=self.mock_snapshot,
            )
            self.assertEqual(2, find_mock.call_count)
